(* This file is the only one you should modify *)
open Ast
open Queues

exception InterpUnimplemented

(*
let mt_heap              = fun x -> 0
let get_var heap str     = heap str
let set_var heap str num = fun x -> if x=str then num else heap x
*)

(* (2b) Implementation of heaps as list *)
let mt_heap = []
let rec get_var heap str = 
    match heap with 
        [] -> 0
      | (var,value)::heap' -> 
            if var=str then value else get_var heap' str
let set_var heap str value = (str,value)::heap

let rec interp_exp h e =
  let interp_exp_r = interp_exp h in
  match e with
    Int i -> i
  | Var v -> get_var h v
  | Plus(e1,e2)  -> (interp_exp_r e1) + (interp_exp_r e2)
  | Times(e1,e2) -> (interp_exp_r e1) * (interp_exp_r e2)
 
let rec interp_step h s =
  let normal h s = (h,s,false,None) in
  match s with
    Skip -> normal h s
  | Assign(v,e) -> normal (set_var h v (interp_exp h e)) Skip
  | Seq(Skip,s2) -> normal h s2
  | Seq(s1,s2) -> 
      let (h3,s3,b,so) = interp_step h s1 in
      (h3, Seq(s3,s2), b, so) (*THIS LINE*)
  | If(e,s1,s2) ->
      normal h (if ((interp_exp h e) <= 0) then s2 else s1)
  | While(e,s1) -> normal h (If(e,Seq(s1,s),Skip))
  | Yield    -> (h,Skip,true,None)
  | Spawn(s) -> (h,Skip,false,Some s)

let rec iter h q s =
  let is_empty q = (try (ignore(deq1 q); false) with DequeueEmpty -> true) in
  match (h,s) with
    (_,Skip) -> 
      if is_empty q 
      then get_var h "ans" 
      else let s,q = deq1 q in iter h q s 
  | _ -> 
      let (h,s,b,so) = interp_step h s in 
      let q = (match so with None -> q | Some s -> enq1 q s) in
      if not b || is_empty q
      then iter h q s
      else
	let s',q = deq1 q in
	iter h (enq1 q s) s' 
  
let interp = iter mt_heap empty_queue1

(* TODO: Write spawn_count1 & 2 *)

(* the next two bindings really belong in a "main" module, but this
keeps all the code you need in one place *)

(* expect one command-line argument, a file to parse and interpret *)
(* you do not need to understand this interaction with the system *)
let get_prog () =
  let argv = Sys.argv in
  let _ = 
    if Array.length argv != 2
    then (prerr_string ("usage: " ^ argv.(0) ^ " [file-to-interpret]\n");
	  exit 1) in
  let ch = open_in argv.(1) in
  Parse.program Lex.lexer (Lexing.from_channel ch)

let _ =
  let prog = get_prog () in
  print_string ("ans = " ^ string_of_int (interp prog) ^"\n")

(* (5a) *)
let rec spawn_count1 s = 
    match s with
        Skip -> 0
      | Assign(_,_) -> 0
      | Yield -> 0
      | Seq(s1,s2) -> (spawn_count1 s1) + (spawn_count1 s2)
      | If (_, s1, s2) -> (spawn_count1 s1) + (spawn_count1 s2)
      | While(_, s) -> (spawn_count1 s)
      | Spawn(s) -> 1 + (spawn_count1 s)

(* (5b) *)
let rec spawn_count2 s =
    match s with
        Skip -> Some 0
      | Assign(_,_) -> Some 0
      | Yield -> Some 0
      | Seq(s1,s2) -> 
            (match (spawn_count2 s1, spawn_count2 s2) with
                (Some c1, Some c2) -> Some (c1 + c2) 
              | _ -> None
            )
      | If (_, s1, s2) -> 
            (match (spawn_count2 s1, spawn_count2 s2) with
                (Some c1, Some c2) -> Some (max c1 c2)
              | _ -> None
            )
      | While(_, s) -> 
            (match spawn_count2 s with
                Some 0 -> Some 0 
              | _ -> None
            )
      | Spawn(s) -> 
            (match spawn_count2 s with
                Some i -> Some (i + 1)
              | _ -> None
            )
